@@ -1,6 +1,6 @@
 ;;; exwm-layout.el --- Layout Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -117,6 +117,8 @@ See variable `exwm-layout-auto-iconify'."
          (width (- (pop edges) x))
          (height (- (pop edges) y))
          frame-x frame-y frame-width frame-height)
+    (when (< emacs-major-version 31)
+      (setq y (+ y (window-tab-line-height window))))
     (with-current-buffer (exwm--id->buffer id)
       (when exwm--floating-frame
         (setq frame-width (frame-pixel-width exwm--floating-frame)
@@ -187,7 +189,6 @@ See variable `exwm-layout-auto-iconify'."
       (exwm-layout--auto-iconify)
       (xcb:flush exwm--connection))))
 
-;;;###autoload
 (cl-defun exwm-layout-set-fullscreen (&optional id)
   "Make window ID fullscreen."
   (interactive)
@@ -213,7 +214,6 @@ See variable `exwm-layout-auto-iconify'."
     (set-window-dedicated-p (get-buffer-window) t)
     (exwm-input--release-keyboard exwm--id)))
 
-;;;###autoload
 (cl-defun exwm-layout-unset-fullscreen (&optional id)
   "Restore X window ID from fullscreen state."
   (interactive)
@@ -244,18 +244,17 @@ See variable `exwm-layout-auto-iconify'."
     (when (eq 'line-mode exwm--selected-input-mode)
       (exwm-input--grab-keyboard exwm--id))))
 
-;;;###autoload
-(cl-defun exwm-layout-toggle-fullscreen (&optional id)
-  "Toggle fullscreen mode of X window ID."
-  (interactive (list (exwm--buffer->id (window-buffer))))
-  (exwm--log "id=#x%x" (or id 0))
-  (unless (or id (derived-mode-p 'exwm-mode))
-    (cl-return-from exwm-layout-toggle-fullscreen))
-  (when id
-    (with-current-buffer (exwm--id->buffer id)
-      (if (exwm-layout--fullscreen-p)
-          (exwm-layout-unset-fullscreen id)
-        (exwm-layout-set-fullscreen id)))))
+(defun exwm-layout-toggle-fullscreen (&optional id)
+  "Toggle fullscreen mode of X window ID.
+If ID is non-nil, default to ID of `window-buffer'."
+  (interactive)
+  (setq id (or id (exwm--buffer->id (current-buffer))
+               (user-error "Current buffer has no X window ID")))
+  (exwm--log "id=#x%x" id)
+  (with-current-buffer (exwm--id->buffer id)
+    (if (exwm-layout--fullscreen-p)
+        (exwm-layout-unset-fullscreen id)
+      (exwm-layout-set-fullscreen id))))
 
 (defun exwm-layout--other-buffer-predicate (buffer)
   "Return non-nil when the BUFFER may be displayed in selected frame.
@@ -385,6 +384,7 @@ If FRAME is nil, refresh layout of selected frame."
                                     (car-safe (window-prev-buffers window)))))
                   (and
                    prev-buffer
+                   (buffer-live-p prev-buffer)
                    (with-current-buffer prev-buffer
                      (derived-mode-p 'exwm-mode))
                    (push prev-buffer covered-buffers)))))))))
@@ -433,7 +433,6 @@ If DIRTY is non-nil, refresh layout immediately."
           (exwm-layout--refresh exwm-workspace--current)
         (exwm--defer 0 #'exwm-layout--refresh exwm-workspace--current)))))
 
-;;;###autoload
 (defun exwm-layout-enlarge-window (delta &optional horizontal)
   "Make the selected window DELTA pixels taller.
 
@@ -523,7 +522,6 @@ windows."
                            :height height))
         (xcb:flush exwm--connection))))))
 
-;;;###autoload
 (defun exwm-layout-enlarge-window-horizontally (delta)
   "Make the selected window DELTA pixels wider.
 
@@ -532,7 +530,6 @@ See also `exwm-layout-enlarge-window'."
   (exwm--log "%s" delta)
   (exwm-layout-enlarge-window delta t))
 
-;;;###autoload
 (defun exwm-layout-shrink-window (delta)
   "Make the selected window DELTA pixels lower.
 
@@ -541,7 +538,6 @@ See also `exwm-layout-enlarge-window'."
   (exwm--log "%s" delta)
   (exwm-layout-enlarge-window (- delta)))
 
-;;;###autoload
 (defun exwm-layout-shrink-window-horizontally (delta)
   "Make the selected window DELTA pixels narrower.
 
@@ -550,7 +546,6 @@ See also `exwm-layout-enlarge-window'."
   (exwm--log "%s" delta)
   (exwm-layout-enlarge-window (- delta) t))
 
-;;;###autoload
 (defun exwm-layout-hide-mode-line ()
   "Hide mode-line."
   (interactive)
@@ -569,7 +564,6 @@ See also `exwm-layout-enlarge-window'."
                              mode-line-height)
                           nil t)))))
 
-;;;###autoload
 (defun exwm-layout-show-mode-line ()
   "Show mode-line."
   (interactive)
@@ -587,7 +581,6 @@ See also `exwm-layout-enlarge-window'."
       (call-interactively #'exwm-input-grab-keyboard))
     (force-mode-line-update)))
 
-;;;###autoload
 (defun exwm-layout-toggle-mode-line ()
   "Toggle the display of mode-line."
   (interactive)
@@ -621,8 +614,5 @@ See also `exwm-layout-enlarge-window'."
     (setq exwm-layout--timer nil))
   (remove-hook 'echo-area-clear-hook #'exwm-layout--on-echo-area-change))
 
-
-
 (provide 'exwm-layout)
-
 ;;; exwm-layout.el ends here
