@@ -1,6 +1,6 @@
 ;;; exwm-randr.el --- RandR Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -30,13 +30,12 @@
 ;; `exwm-randr-workspace-monitor-plist' and `exwm-randr-screen-change-hook'
 ;; as follows:
 ;;
-;;   (require 'exwm-randr)
 ;;   (setq exwm-randr-workspace-monitor-plist '(0 "VGA1"))
 ;;   (add-hook 'exwm-randr-screen-change-hook
 ;;             (lambda ()
 ;;               (start-process-shell-command
 ;;                "xrandr" nil "xrandr --output VGA1 --left-of LVDS1 --auto")))
-;;   (exwm-randr-enable)
+;;   (exwm-randr-mode 1)
 ;;
 ;; With above lines, workspace 0 should be assigned to the output named "VGA1",
 ;; staying at the left of other workspaces on the output "LVDS1".  Please refer
@@ -58,6 +57,8 @@
   "RandR."
   :group 'exwm)
 
+(defvar exwm-randr--connection nil "The X connection.")
+
 (defcustom exwm-randr-refresh-hook nil
   "Normal hook run when the RandR module just refreshed."
   :type 'hook)
@@ -87,23 +88,48 @@ setting workspace other than 1 and 3 would always be displayed on the
 primary monitor where workspace 1 and 3 would be displayed on their
 corresponding monitors whenever the monitors are active.
 
+Changes to this variable only take immediate affect when set before
+`exwm-randr-mode' is enabled, via `setopt', or when customized (see the
+Info node `Customization'). Otherwise, the `exwm-randr-refresh' must be
+called explicitly to assign the correct workspaces to the correct monitors.
+
   \\='(1 \"HDMI-1\" 3 \"DP-1\")"
-  :type '(plist :key-type integer :value-type string))
+  :type '(plist :key-type integer :value-type string)
+  :initialize 'custom-initialize-changed
+  :set (lambda (symbol value)
+         (set-default-toplevel-value symbol value)
+         (when exwm-randr--connection
+           (exwm-randr-refresh))))
+
+(defvar exwm-randr--connection nil "The X connection.")
 
 (defvar exwm-randr--last-timestamp 0 "Used for debouncing events.")
 
-(defvar exwm-randr--prev-screen-change-seqnum nil
-  "The most recent ScreenChangeNotify sequence number.")
+(defvar exwm-randr--prev-screen-change-timestamp 0
+  "The most recent ScreenChangeNotify config change timestamp.")
 
 (defvar exwm-randr--compatibility-mode nil
   "Non-nil when the server does not support RandR 1.5 protocol.")
 
+;;;###autoload
+(define-minor-mode exwm-randr-mode
+  "Toggle EXWM randr support."
+  :global t
+  :group 'exwm-randr
+  (exwm--global-minor-mode-body randr))
+
+(defsubst exwm-randr--assert-connected ()
+  "Assert that `exwm-randr-mode' is enabled and activated."
+  (cond
+   ((not exwm-randr-mode) (user-error "EXWM RandR mode not enabled"))
+   ((not exwm-randr--connection) (user-error "EXWM RandR not connected, is EXWM running?"))))
+
 (defun exwm-randr--get-monitors ()
   "Get RandR 1.5 monitors."
   (exwm--log)
   (let (monitor-name geometry monitor-geometry-alist primary-monitor)
     (with-slots (timestamp monitors)
-        (xcb:+request-unchecked+reply exwm--connection
+        (xcb:+request-unchecked+reply exwm-randr--connection
             (make-instance 'xcb:randr:GetMonitors
                            :window exwm--root
                            :get-active 1))
@@ -136,21 +162,21 @@ Only used when RandR 1.5 is not supported by the server."
   (exwm--log)
   (let (output-name geometry output-geometry-alist primary-output)
     (with-slots (config-timestamp outputs)
-        (xcb:+request-unchecked+reply exwm--connection
+        (xcb:+request-unchecked+reply exwm-randr--connection
             (make-instance 'xcb:randr:GetScreenResourcesCurrent
                            :window exwm--root))
       (when (> config-timestamp exwm-randr--last-timestamp)
         (setq exwm-randr--last-timestamp config-timestamp))
       (dolist (output outputs)
         (with-slots (crtc connection name)
-            (xcb:+request-unchecked+reply exwm--connection
+            (xcb:+request-unchecked+reply exwm-randr--connection
                 (make-instance 'xcb:randr:GetOutputInfo
                                :output output
                                :config-timestamp config-timestamp))
           (when (and (= connection xcb:randr:Connection:Connected)
                      (/= crtc 0))
             (with-slots (x y width height)
-                (xcb:+request-unchecked+reply exwm--connection
+                (xcb:+request-unchecked+reply exwm-randr--connection
                     (make-instance 'xcb:randr:GetCrtcInfo
                                    :crtc crtc
                                    :config-timestamp config-timestamp))
@@ -198,11 +224,11 @@ In a mirroring setup some monitors overlap and should be treated as one."
                                             monitor-alias-alist))))))
     monitor-alias-alist))
 
-;;;###autoload
 (defun exwm-randr-refresh ()
   "Refresh workspaces according to the updated RandR info."
   (interactive)
   (exwm--log)
+  (exwm-randr--assert-connected)
   (let* ((result (if exwm-randr--compatibility-mode
                      (exwm-randr--get-outputs)
                    (exwm-randr--get-monitors)))
@@ -238,7 +264,7 @@ In a mirroring setup some monitors overlap and should be treated as one."
       ;; Resize workspace.
       (dolist (f exwm-workspace--list)
         (exwm-workspace--set-fullscreen f))
-      (xcb:flush exwm--connection)
+      (xcb:flush exwm-randr--connection)
       ;; Raise the minibuffer if it's active.
       (when (and (active-minibuffer-window)
                  (exwm-workspace--minibuffer-own-frame-p))
@@ -251,7 +277,7 @@ In a mirroring setup some monitors overlap and should be treated as one."
       ;; Mark the workspace on the top of each monitor as active.
       (dolist (xwin
                (reverse
-                (slot-value (xcb:+request-unchecked+reply exwm--connection
+                (slot-value (xcb:+request-unchecked+reply exwm-randr--connection
                                 (make-instance 'xcb:QueryTree
                                                :window exwm--root))
                             'children)))
@@ -261,23 +287,23 @@ In a mirroring setup some monitors overlap and should be treated as one."
                   (rassq-delete-all monitor container-monitor-alist))
             (exwm-workspace--set-active (cdr (assq xwin container-frame-alist))
                                         t))))
-      (xcb:flush exwm--connection)
+      (xcb:flush exwm-randr--connection)
       (run-hooks 'exwm-randr-refresh-hook))))
 
 (defun exwm-randr--on-ScreenChangeNotify (data _synthetic)
-  "Handle `ScreenChangeNotify' event.
+  "Handle `ScreenChangeNotify' event with DATA.
 
 Run `exwm-randr-screen-change-hook' (usually user scripts to configure RandR)."
   (exwm--log)
   (let ((evt (make-instance 'xcb:randr:ScreenChangeNotify)))
     (xcb:unmarshal evt data)
-    (let ((seqnum (slot-value evt '~sequence)))
-      (unless (equal seqnum exwm-randr--prev-screen-change-seqnum)
-        (setq exwm-randr--prev-screen-change-seqnum seqnum)
+    (let ((ts (slot-value evt 'config-timestamp)))
+      (unless (equal ts exwm-randr--prev-screen-change-timestamp)
+        (setq exwm-randr--prev-screen-change-timestamp ts)
         (run-hooks 'exwm-randr-screen-change-hook)))))
 
 (defun exwm-randr--on-Notify (data _synthetic)
-  "Handle `CrtcChangeNotify' and `OutputChangeNotify' events.
+  "Handle `CrtcChangeNotify' and `OutputChangeNotify' events with DATA.
 
 Refresh when any CRTC/output changes."
   (exwm--log)
@@ -285,11 +311,10 @@ Refresh when any CRTC/output changes."
         notify)
     (xcb:unmarshal evt data)
     (with-slots (subCode u) evt
-      (cl-case subCode
-        (xcb:randr:Notify:CrtcChange
-         (setq notify (slot-value u 'cc)))
-        (xcb:randr:Notify:OutputChange
-         (setq notify (slot-value u 'oc))))
+      (cond ((= subCode xcb:randr:Notify:CrtcChange)
+             (setq notify (slot-value u 'cc)))
+            ((= subCode xcb:randr:Notify:OutputChange)
+             (setq notify (slot-value u 'oc))))
       (when notify
         (with-slots (timestamp) notify
           (when (> timestamp exwm-randr--last-timestamp)
@@ -297,7 +322,7 @@ Refresh when any CRTC/output changes."
             (setq exwm-randr--last-timestamp timestamp)))))))
 
 (defun exwm-randr--on-ConfigureNotify (data _synthetic)
-  "Handle `ConfigureNotify' event.
+  "Handle `ConfigureNotify' event with DATA.
 
 Refresh when any RandR 1.5 monitor changes."
   (exwm--log)
@@ -307,14 +332,20 @@ Refresh when any RandR 1.5 monitor changes."
       (when (eq window exwm--root)
         (exwm-randr-refresh)))))
 
-(defun exwm-randr--init ()
+(cl-defun exwm-randr--init ()
   "Initialize RandR extension and EXWM RandR module."
   (exwm--log)
-  (when (= 0 (slot-value (xcb:get-extension-data exwm--connection 'xcb:randr)
+  (when exwm-randr--connection
+    (cl-return-from exwm-randr--init))
+  (setq exwm-randr--connection (xcb:connect))
+  (set-process-query-on-exit-flag (slot-value exwm-randr--connection 'process) nil)
+  (when (= 0 (slot-value (xcb:get-extension-data exwm-randr--connection 'xcb:randr)
                          'present))
+    (xcb:disconnect exwm-randr--connection)
+    (setq exwm-randr--connection nil)
     (error "[EXWM] RandR extension is not supported by the server"))
   (with-slots (major-version minor-version)
-      (xcb:+request-unchecked+reply exwm--connection
+      (xcb:+request-unchecked+reply exwm-randr--connection
           (make-instance 'xcb:randr:QueryVersion
                          :major-version 1 :minor-version 5))
     (cond ((and (= major-version 1) (= minor-version 5))
@@ -322,6 +353,8 @@ Refresh when any RandR 1.5 monitor changes."
           ((and (= major-version 1) (>= minor-version 2))
            (setq exwm-randr--compatibility-mode t))
           (t
+           (xcb:disconnect exwm-randr--connection)
+           (setq exwm-randr--connection nil)
            (error "[EXWM] The server only support RandR version up to %d.%d"
                   major-version minor-version)))
     ;; External monitor(s) may already be connected.
@@ -330,20 +363,20 @@ Refresh when any RandR 1.5 monitor changes."
     ;; Listen for `ScreenChangeNotify' to notify external tools to
     ;; configure RandR and `CrtcChangeNotify/OutputChangeNotify' to
     ;; refresh the workspace layout.
-    (xcb:+event exwm--connection 'xcb:randr:ScreenChangeNotify
+    (xcb:+event exwm-randr--connection 'xcb:randr:ScreenChangeNotify
                 #'exwm-randr--on-ScreenChangeNotify)
-    (xcb:+event exwm--connection 'xcb:randr:Notify
+    (xcb:+event exwm-randr--connection 'xcb:randr:Notify
                 #'exwm-randr--on-Notify)
-    (xcb:+event exwm--connection 'xcb:ConfigureNotify
+    (xcb:+event exwm-randr--connection 'xcb:ConfigureNotify
                 #'exwm-randr--on-ConfigureNotify)
-    (xcb:+request exwm--connection
+    (xcb:+request exwm-randr--connection
         (make-instance 'xcb:randr:SelectInput
                        :window exwm--root
                        :enable (logior
                                 xcb:randr:NotifyMask:ScreenChange
                                 xcb:randr:NotifyMask:CrtcChange
                                 xcb:randr:NotifyMask:OutputChange)))
-    (xcb:flush exwm--connection)
+    (xcb:flush exwm-randr--connection)
     (add-hook 'exwm-workspace-list-change-hook #'exwm-randr-refresh))
   ;; Prevent frame parameters introduced by this module from being
   ;; saved/restored.
@@ -354,16 +387,10 @@ Refresh when any RandR 1.5 monitor changes."
 (defun exwm-randr--exit ()
   "Exit the RandR module."
   (exwm--log)
-  (remove-hook 'exwm-workspace-list-change-hook #'exwm-randr-refresh))
-
-(defun exwm-randr-enable ()
-  "Enable RandR support for EXWM."
-  (exwm--log)
-  (add-hook 'exwm-init-hook #'exwm-randr--init)
-  (add-hook 'exwm-exit-hook #'exwm-randr--exit))
-
-
+  (remove-hook 'exwm-workspace-list-change-hook #'exwm-randr-refresh)
+  (when exwm-randr--connection
+    (xcb:disconnect exwm-randr--connection)
+    (setq exwm-randr--connection nil)))
 
 (provide 'exwm-randr)
-
 ;;; exwm-randr.el ends here
