@@ -1,7 +1,7 @@
 ;;; exwm-systemtray.el --- System Tray Module for  -*- lexical-binding: t -*-
 ;;;                        EXWM
 
-;; Copyright (C) 2016-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2016-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -57,12 +57,6 @@
   :group 'exwm-systemtray
   (exwm--global-minor-mode-body systemtray))
 
-(defun exwm-systemtray-enable ()
-  "Enable EXWM systemtray support."
-  (message "`exwm-systemtray-enable' is obsolete. Use `exwm-systemtray-mode' instead.")
-  (exwm-systemtray-mode 1))
-(make-obsolete 'exwm-systemtray-enable "Use `exwm-systemtray-mode' instead." "0.30")
-
 (defcustom exwm-systemtray-height nil
   "System tray height.
 
@@ -695,8 +689,5 @@ Argument DATA contains the raw event data."
     (when (boundp 'exwm-randr-refresh-hook)
       (remove-hook 'exwm-randr-refresh-hook #'exwm-systemtray--refresh-all))))
 
-
-
 (provide 'exwm-systemtray)
-
 ;;; exwm-systemtray.el ends here
