@@ -1,6 +1,6 @@
 ;;; exwm-workspace.el --- Workspace Module for EXWM  -*- lexical-binding: t -*-
 
-;; Copyright (C) 1015-2024 Free Software Foundation, Inc.
+;; Copyright (C) 1015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 
@@ -54,11 +54,35 @@ This happens when a workspace is added, deleted, moved, etc."
   "Initial number of workspaces."
   :type 'integer)
 
+(defvar exwm-workspace--switch-history nil
+  "History for `read-from-minibuffer' to interactively switch workspace.")
+
+(defvar exwm-workspace--switch-history-outdated nil
+  "Non-nil to indicate `exwm-workspace--switch-history' is outdated.")
+
+(defvar exwm-workspace--switch-by-index-map
+  (let ((map (make-sparse-keymap)))
+    (dotimes (i 10)
+      (keymap-set map (int-to-string i)
+                  #'exwm-workspace--switch-map-nth-prefix))
+    map)
+  "Keymap used to switch workspaces by index.")
+
+(defvar exwm-workspace--switch-by-name-map (make-sparse-keymap)
+  "Keymap used to switch workspaces by index.
+Applicable when the name returned by `exwm-workspace-index-map' is also
+a valid key.")
+
 (defcustom exwm-workspace-index-map #'number-to-string
   "Function for mapping a workspace index to a string for display.
 
 By default `number-to-string' is applied which yields 0 1 2 ... ."
-  :type 'function)
+  :type 'function
+  :initialize 'custom-initialize-changed
+  :set (lambda (symbol value)
+         (set-default-toplevel-value symbol value)
+         (setq exwm-workspace--switch-history-outdated t)
+         (exwm-workspace--update-switch-by-name-map)))
 
 (defcustom exwm-workspace-minibuffer-position nil
   "Position of the minibuffer frame.
@@ -116,12 +140,6 @@ Please manually run the hook `exwm-workspace-list-change-hook' afterwards.")
 
 (defvar exwm-workspace--struts nil "Areas occupied by struts.")
 
-(defvar exwm-workspace--switch-history nil
-  "History for `read-from-minibuffer' to interactively switch workspace.")
-
-(defvar exwm-workspace--switch-history-outdated nil
-  "Non-nil to indicate `exwm-workspace--switch-history' is outdated.")
-
 (defvar exwm-workspace--timer nil "Timer used to track echo area changes.")
 
 (defvar exwm-workspace--update-workareas-hook nil
@@ -134,7 +152,6 @@ Please manually run the hook `exwm-workspace-list-change-hook' afterwards.")
 (defvar exwm-workspace--window-y-offset 0
   "Offset between Emacs first window & outer frame in Y.")
 
-(defvar exwm-input--during-command)
 (defvar exwm-input--event-hook)
 (defvar exwm-layout-show-all-buffers)
 (defvar exwm-manage--desktop)
@@ -170,45 +187,30 @@ FRAME may be either a workspace frame or a workspace position."
            frame
          (exwm-workspace--position frame))))
 
-(defvar exwm-workspace--switch-map nil
-  "Keymap used for interactively selecting workspace.")
+(defvar-keymap exwm-workspace-switch-map
+  :doc "Keymap used by `exwm-workspace-switch'."
+  "+" #'exwm-workspace--prompt-add
+  "-" #'exwm-workspace--prompt-delete
+  "C-a" (lambda () (interactive) (goto-history-element 1))
+  "C-e" (lambda () (interactive) (goto-history-element (exwm-workspace--count)))
+  "C-g" #'abort-recursive-edit
+  "C-]" #'abort-recursive-edit
+  "C-j" #'exit-minibuffer
+  "<return>" #'exit-minibuffer
+  "<space>" #'exit-minibuffer
+  "C-f" #'previous-history-element
+  "C-b" #'next-history-element
+  ;; Alternative keys
+  "<right>" #'previous-history-element
+  "<left>" #'next-history-element)
 
-(defun exwm-workspace--init-switch-map ()
-  "Initialize variable `exwm-workspace--switch-map'."
-  (let ((map (make-sparse-keymap)))
-    (define-key map [t] (lambda () (interactive)))
-    (define-key map "+" #'exwm-workspace--prompt-add)
-    (define-key map "-" #'exwm-workspace--prompt-delete)
-    (dotimes (i 10)
-      (define-key map (int-to-string i)
-        #'exwm-workspace--switch-map-nth-prefix))
-    (unless (eq exwm-workspace-index-map #'number-to-string)
-      ;; Add extra (and possibly override) keys for selecting workspace.
-      (dotimes (i 10)
-        (let ((key (funcall exwm-workspace-index-map i)))
-          (when (and (stringp key)
-                     (= (length key) 1)
-                     (<= 0 (elt key 0) 127))
-            (define-key map key
-              (lambda ()
-                (interactive)
-                (exwm-workspace--switch-map-select-nth i)))))))
-    (define-key map "\C-a" (lambda () (interactive) (goto-history-element 1)))
-    (define-key map "\C-e" (lambda ()
-                             (interactive)
-                             (goto-history-element (exwm-workspace--count))))
-    (define-key map "\C-g" #'abort-recursive-edit)
-    (define-key map "\C-]" #'abort-recursive-edit)
-    (define-key map "\C-j" #'exit-minibuffer)
-    ;; (define-key map "\C-m" #'exit-minibuffer) ;not working
-    (define-key map [return] #'exit-minibuffer)
-    (define-key map " " #'exit-minibuffer)
-    (define-key map "\C-f" #'previous-history-element)
-    (define-key map "\C-b" #'next-history-element)
-    ;; Alternative keys
-    (define-key map [right] #'previous-history-element)
-    (define-key map [left] #'next-history-element)
-    (setq exwm-workspace--switch-map map)))
+(defvar exwm-workspace--switch-composed-map
+  (make-composed-keymap (list
+                         exwm-workspace-switch-map
+                         exwm-workspace--switch-by-name-map
+                         exwm-workspace--switch-by-index-map)
+                        (define-keymap "<t>" #'undefined))
+  "Internal Keymap composing all the keymaps used by `exwm-workspace-switch'.")
 
 (defun exwm-workspace--workspace-from-frame-or-index (frame-or-index)
   "Retrieve the workspace frame from FRAME-OR-INDEX."
@@ -233,7 +235,7 @@ Show PROMPT to the user if non-nil."
          (history-idx (read-from-minibuffer
                        (or prompt "Workspace: ")
                        (elt exwm-workspace--switch-history current-idx)
-                       exwm-workspace--switch-map nil
+                       exwm-workspace--switch-composed-map nil
                        `(exwm-workspace--switch-history . ,(1+ current-idx))))
          (workspace-idx (cl-position history-idx exwm-workspace--switch-history
                                      :test #'equal)))
@@ -295,7 +297,20 @@ Show PROMPT to the user if non-nil."
                 sequence ""))
              sequence)))))
 
-;;;###autoload
+(defun exwm-workspace--update-switch-by-name-map ()
+  "Updates `exwm-workspace--switch-by-name-map'."
+  (setcdr exwm-workspace--switch-by-name-map nil)
+  (unless (eq exwm-workspace-index-map #'number-to-string)
+    (dotimes (i 10)
+      (let ((key (funcall exwm-workspace-index-map i)))
+        (when (and (stringp key)
+                   (length= key 1)
+                   (<= 0 (elt key 0) 127))
+          (keymap-set exwm-workspace--switch-by-name-map key
+                      (lambda ()
+                        (interactive)
+                        (exwm-workspace--switch-map-select-nth i))))))))
+
 (defun exwm-workspace--get-geometry (frame)
   "Return the geometry of frame FRAME."
   (or (frame-parameter frame 'exwm-geometry)
@@ -305,7 +320,6 @@ Show PROMPT to the user if non-nil."
                      :width (x-display-pixel-width)
                      :height (x-display-pixel-height))))
 
-;;;###autoload
 (defun exwm-workspace--current-height ()
   "Return the height of current workspace."
   (let ((geometry (frame-parameter exwm-workspace--current 'exwm-geometry)))
@@ -313,7 +327,6 @@ Show PROMPT to the user if non-nil."
         (slot-value geometry 'height)
       (x-display-pixel-height))))
 
-;;;###autoload
 (defun exwm-workspace--minibuffer-own-frame-p ()
   "Reports whether the minibuffer is displayed in its own frame."
   (memq exwm-workspace-minibuffer-position '(top bottom)))
@@ -346,7 +359,7 @@ Show PROMPT to the user if non-nil."
          ;; Get workareas prior to struts.
          (workareas (mapcar
                      (lambda (frame)
-                       (if-let (rect (frame-parameter frame 'exwm-geometry))
+                       (if-let* ((rect (frame-parameter frame 'exwm-geometry)))
                            ;; Use the 'exwm-geometry' frame parameter if it
                            ;; exists.  Make sure to clone it, will be modified
                            ;; below!
@@ -547,9 +560,9 @@ PREFIX-DIGITS is a list of the digits introduced so far."
                      (interactive)
                      (exwm-workspace--switch-map-nth-prefix digits)))))
          (dotimes (i 10)
-           (define-key map (int-to-string i) cmd))
+           (keymap-set map (int-to-string i) cmd))
          ;; Accept
-         (define-key map [return]
+         (keymap-set map "<return>"
            (lambda ()
              (interactive)
              (exwm-workspace--switch-map-select-nth n)))
@@ -561,7 +574,6 @@ PREFIX-DIGITS is a list of the digits introduced so far."
   (goto-history-element (1+ n))
   (exit-minibuffer))
 
-;;;###autoload
 (defun exwm-workspace-switch (frame-or-index &optional force)
   "Switch to workspace FRAME-OR-INDEX (0-based).
 
@@ -693,7 +705,6 @@ When FORCE is true, allow switching to current workspace."
     (funcall exwm-workspace--original-handle-focus-in (list 'focus-in frame))
     (run-hooks 'exwm-workspace-switch-hook)))
 
-;;;###autoload
 (defun exwm-workspace-switch-create (frame-or-index)
   "Switch to workspace FRAME-OR-INDEX creating it first non-existent.
 
@@ -718,7 +729,6 @@ Passing a workspace frame as the first option is for internal use only."
       (run-hooks 'exwm-workspace-list-change-hook))
     (exwm-workspace-switch frame-or-index)))
 
-;;;###autoload
 (defun exwm-workspace-swap (workspace1 workspace2)
   "Interchange position of WORKSPACE1 with that of WORKSPACE2."
   (interactive
@@ -754,7 +764,6 @@ Passing a workspace frame as the first option is for internal use only."
         (exwm-workspace-switch exwm-workspace--current t))
       (run-hooks 'exwm-workspace-list-change-hook))))
 
-;;;###autoload
 (defun exwm-workspace-move (workspace nth)
   "Move WORKSPACE to the NTH position.
 
@@ -801,7 +810,6 @@ before it."
         (exwm-workspace-switch exwm-workspace--current t))
       (run-hooks 'exwm-workspace-list-change-hook))))
 
-;;;###autoload
 (defun exwm-workspace-add (&optional index)
   "Add a workspace as the INDEX-th workspace, or the last one if INDEX is nil.
 
@@ -814,7 +822,6 @@ INDEX must not exceed the current number of workspaces."
       (exwm-workspace-move (make-frame) index)
     (make-frame)))
 
-;;;###autoload
 (defun exwm-workspace-delete (&optional frame-or-index)
   "Delete the workspace FRAME-OR-INDEX."
   (interactive)
@@ -842,7 +849,6 @@ INDEX must not exceed the current number of workspaces."
                          :window id
                          :data desktop)))))
 
-;;;###autoload
 (cl-defun exwm-workspace-move-window (frame-or-index &optional id)
   "Move window ID to workspace FRAME-OR-INDEX."
   (interactive (list
@@ -984,7 +990,6 @@ INDEX must not exceed the current number of workspaces."
         (xcb:flush exwm--connection)))
     (setq exwm-workspace--switch-history-outdated t)))
 
-;;;###autoload
 (defun exwm-workspace-switch-to-buffer (buffer-or-name)
   "Make selected window display BUFFER-OR-NAME."
   (interactive
@@ -1076,7 +1081,6 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
   (assq (frame-parameter exwm-workspace--minibuffer 'exwm-container)
         exwm-workspace--id-struts-alist))
 
-;;;###autoload
 (defun exwm-workspace-attach-minibuffer ()
   "Attach the minibuffer making it always visible."
   (interactive)
@@ -1101,7 +1105,6 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
       (dolist (f exwm-workspace--list)
         (exwm-workspace--set-fullscreen f)))))
 
-;;;###autoload
 (defun exwm-workspace-detach-minibuffer ()
   "Detach the minibuffer so that it automatically hides."
   (interactive)
@@ -1119,7 +1122,6 @@ Please check `exwm-workspace--minibuffer-own-frame-p' first."
         (exwm-workspace--set-fullscreen f))
       (exwm-workspace--hide-minibuffer))))
 
-;;;###autoload
 (defun exwm-workspace-toggle-minibuffer ()
   "Attach the minibuffer if it's detached, or detach it if it's attached."
   (interactive)
@@ -1284,7 +1286,7 @@ ALIST is an action alist, as accepted by function `display-buffer'."
     (exwm-workspace--update-minibuffer-height t)
     (exwm-workspace--show-minibuffer)
     (unless (or (not exwm-workspace-display-echo-area-timeout)
-                exwm-input--during-command ;e.g. read-event
+                real-this-command ;e.g. read-event
                 input-method-use-echo-area)
       (setq exwm-workspace--display-echo-area-timer
             (run-with-timer exwm-workspace-display-echo-area-timeout nil
@@ -1665,7 +1667,10 @@ applied to all subsequently created X frames."
 (defun exwm-workspace--init ()
   "Initialize workspace module."
   (exwm--log)
-  (exwm-workspace--init-switch-map)
+  ;; Re-initialize the workspace switch-by-name map just in case
+  ;; the user customized it via setq instead of setopt. This preserves
+  ;; existing behavior and doesn't really add any complexity.
+  (exwm-workspace--update-switch-by-name-map)
   ;; Prevent unexpected exit
   (setq exwm-workspace--fullscreen-frame-count 0)
   (exwm-workspace--modify-all-x-frames-parameters
@@ -1764,8 +1769,5 @@ applied to all subsequently created X frames."
       (accept-process-output nil 0.1)))
   (setq exwm-workspace--fullscreen-frame-count nil))
 
-
-
 (provide 'exwm-workspace)
-
 ;;; exwm-workspace.el ends here
