@@ -1,6 +1,6 @@
 ;;; exwm-xsettings.el --- XSETTINGS Module for EXWM -*- lexical-binding: t -*-
 
-;; Copyright (C) 2022-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2022-2025 Free Software Foundation, Inc.
 
 ;; Author: Steven Allen <steven@stebalien.com>
 
@@ -68,12 +68,6 @@
   :group 'exwm-xsettings
   (exwm--global-minor-mode-body xsettings))
 
-(defun exwm-xsettings-enable ()
-  "Enable EXWM xsettings support."
-  (message "`exwm-xsettings-enable' is obsolete. Use `exwm-xsettings-mode' instead.")
-  (exwm-xsettings-mode 1))
-(make-obsolete 'exwm-xsettings-enable "Use `exwm-xsettings-mode' instead." "0.30")
-
 (defun exwm-xsettings--rgba-match (_widget value)
   "Return t if VALUE is a valid RGBA color."
   (and (numberp value) (<= 0 value 1)))
@@ -326,5 +320,4 @@ SERIAL is a sequence number."
           exwm-xsettings--selection-owner-window nil)))
 
 (provide 'exwm-xsettings)
-
 ;;; exwm-xsettings.el ends here
