@@ -1,11 +1,11 @@
 ;;; exwm.el --- Emacs X Window Manager  -*- lexical-binding: t -*-
 
-;; Copyright (C) 2015-2024 Free Software Foundation, Inc.
+;; Copyright (C) 2015-2025 Free Software Foundation, Inc.
 
 ;; Author: Chris Feng <chris.w.feng@gmail.com>
 ;; Maintainer: Adrián Medraño Calvo <adrian@medranocalvo.com>, Steven Allen <steven@stebalien.com>, Daniel Mendler <mail@daniel-mendler.de>
-;; Version: 0.30
-;; Package-Requires: ((emacs "27.1") (xelb "0.19"))
+;; Version: 0.33
+;; Package-Requires: ((emacs "27.1") (xelb "0.20") (compat "30"))
 ;; Keywords: unix
 ;; URL: https://github.com/emacs-exwm/exwm
 
@@ -49,10 +49,13 @@
 ;; 2. In '~/.emacs', add following lines (please modify accordingly):
 ;;
 ;;    (require 'exwm)
-;;    (require 'exwm-config)
-;;    (exwm-config-example)
+;;    (setq exwm-input-global-keys `(([?\s-r] . exwm-reset)))
+;;    (exwm-enable)
+;;
+;; 3. Add the following lines to '~/.xinitrc':
+;;
+;;    exec emacs
 ;;
-;; 3. Link or copy the file 'xinitrc' to '~/.xinitrc'.
 ;; 4. Launch EXWM in a console (e.g. tty1) with
 ;;
 ;;    xinit -- vt01
@@ -122,6 +125,10 @@ After this time, the server will be killed.")
 
 (defvar exwm--server-process nil "Process of the subordinate Emacs server.")
 
+(defvar exwm--client-message-functions nil
+  "Alist of form ((MESSAGE . MESSAGE-HANDLER)...).
+Set during `exwm-init'.")
+
 (defun exwm-reset ()
   "Reset the state of the selected window (non-fullscreen, line-mode, etc)."
   (interactive)
@@ -134,12 +141,11 @@ After this time, the server will be killed.")
       (exwm-layout--refresh)
       (call-interactively #'exwm-input-grab-keyboard))))
 
-;;;###autoload
 (defun exwm-restart ()
   "Restart EXWM."
   (interactive)
   (exwm--log)
-  (when (exwm--confirm-kill-emacs "Restart?" 'no-check)
+  (when (exwm--confirm-kill-emacs "Restart" 'no-check)
     (let* ((attr (process-attributes (emacs-pid)))
            (args (cdr (assq 'args attr)))
            (ppid (cdr (assq 'ppid attr)))
@@ -463,188 +469,202 @@ DATA contains unmarshalled PropertyNotify event data."
                           (x-get-atom-name atom exwm-workspace--current)
                           atom)))))))
 
+(defun exwm--on-net-number-of-desktops (_id data)
+  "Handle _NET_NUMBER_OF_DESKTOPS_ message with DATA."
+  (let ((current (exwm-workspace--count))
+        (requested (elt data 0)))
+    ;; Only allow increasing/decreasing the workspace number by 1.
+    (cond
+     ((< current requested)
+      (make-frame))
+     ((and (> current requested)
+           (> current 1))
+      (let ((frame (car (last exwm-workspace--list))))
+        (delete-frame frame))))))
+
+(defun exwm--on-net-current-desktop (_id data)
+  "Handle _NET_CURRENT_DESKTOP message with DATA."
+  (exwm-workspace-switch (elt data 0)))
+
+(defun exwm--on-net-active-window (id _data)
+  "Handle _NET_ACTIVE_WINDOW message with ID."
+  (let ((buffer (exwm--id->buffer id))
+        window)
+    (if (buffer-live-p buffer)
+        ;; Either an `exwm-mode' buffer (an X window) or a floating frame.
+        (with-current-buffer buffer
+          (when (eq exwm--frame exwm-workspace--current)
+            (if exwm--floating-frame
+                (select-frame exwm--floating-frame)
+              (setq window (get-buffer-window nil t))
+              (unless window
+                ;; State change: iconic => normal.
+                (setq window (frame-selected-window exwm--frame))
+                (set-window-buffer window (current-buffer)))
+              ;; Focus transfer.
+              (select-window window))))
+      ;; A workspace.
+      (dolist (f exwm-workspace--list)
+        (when (eq id (frame-parameter f 'exwm-outer-id))
+          (x-focus-frame f t))))))
+
+(defun exwm--on-net-close-window (id _data)
+  "Handle _NET_CLOSE_WINDOW message with ID."
+  (let ((buffer (exwm--id->buffer id)))
+    (when (buffer-live-p buffer)
+      (exwm--defer 0 #'kill-buffer buffer))))
+
+(defun exwm--on-net-wm-moveresize (id data)
+  "Handle _NET_WM_MOVERESIZE message with ID and DATA."
+  (let ((direction (elt data 2))
+        (buffer (exwm--id->buffer id)))
+    (unless (and buffer
+                 (not (buffer-local-value 'exwm--floating-frame buffer)))
+      (cond ((= direction
+                xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_KEYBOARD)
+             ;; FIXME
+             )
+            ((= direction
+                xcb:ewmh:_NET_WM_MOVERESIZE_MOVE_KEYBOARD)
+             ;; FIXME
+             )
+            ((= direction xcb:ewmh:_NET_WM_MOVERESIZE_CANCEL)
+             (exwm-floating--stop-moveresize))
+            ;; In case it's a workspace frame.
+            ((and (not buffer)
+                  (catch 'break
+                    (dolist (f exwm-workspace--list)
+                      (when (or (eq id (frame-parameter f 'exwm-outer-id))
+                                (eq id (frame-parameter f 'exwm-id)))
+                        (throw 'break t)))
+                    nil)))
+            (t
+             ;; In case it's a floating frame,
+             ;; move the corresponding X window instead.
+             (unless buffer
+               (catch 'break
+                 (dolist (pair exwm--id-buffer-alist)
+                   (with-current-buffer (cdr pair)
+                     (when
+                         (and exwm--floating-frame
+                              (or (eq id
+                                      (frame-parameter exwm--floating-frame
+                                                       'exwm-outer-id))
+                                  (eq id
+                                      (frame-parameter exwm--floating-frame
+                                                       'exwm-id))))
+                       (setq id exwm--id)
+                       (throw 'break nil))))))
+             ;; Start to move it.
+             (exwm-floating--start-moveresize id direction))))))
+
+(defun exwm--on-net-request-frame-extents (id _data)
+  "Handle _NET_REQUEST_FRAME_EXTENTS message with ID."
+  (let ((buffer (exwm--id->buffer id))
+        top btm)
+    (if (or (not buffer)
+            (not (buffer-local-value 'exwm--floating-frame buffer)))
+        (setq top 0
+              btm 0)
+      (setq top (window-header-line-height)
+            btm (window-mode-line-height)))
+    (xcb:+request exwm--connection
+        (make-instance 'xcb:ewmh:set-_NET_FRAME_EXTENTS
+                       :window id
+                       :left 0
+                       :right 0
+                       :top top
+                       :bottom btm)))
+  (xcb:flush exwm--connection))
+
+(defun exwm--on-net-wm-desktop (id data)
+  "Handle _NET_WM_DESKTOP message with ID and DATA."
+  (let ((buffer (exwm--id->buffer id)))
+    (when (buffer-live-p buffer)
+      (exwm-workspace-move-window (elt data 0) id))))
+
+(defun exwm--on-net-wm-state (id data)
+  "Handle _NET_WM_STATE message with ID and DATA."
+  (let ((action (elt data 0))
+        (props (list (elt data 1) (elt data 2)))
+        (buffer (exwm--id->buffer id))
+        props-new)
+    ;; only support _NET_WM_STATE_FULLSCREEN / _NET_WM_STATE_ADD for frames
+    (when (and (not buffer)
+               (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
+               (= action xcb:ewmh:_NET_WM_STATE_ADD))
+      (xcb:+request
+          exwm--connection
+          (make-instance 'xcb:ewmh:set-_NET_WM_STATE
+                         :window id
+                         :data (vector xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
+      (xcb:flush exwm--connection))
+    (when buffer                    ;ensure it's managed
+      (with-current-buffer buffer
+        ;; _NET_WM_STATE_FULLSCREEN
+        (when (or (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
+                  (memq xcb:Atom:_NET_WM_STATE_ABOVE props))
+          (cond ((= action xcb:ewmh:_NET_WM_STATE_ADD)
+                 (unless (exwm-layout--fullscreen-p)
+                   (exwm-layout-set-fullscreen id))
+                 (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new))
+                ((= action xcb:ewmh:_NET_WM_STATE_REMOVE)
+                 (when (exwm-layout--fullscreen-p)
+                   (exwm-layout-unset-fullscreen id)))
+                ((= action xcb:ewmh:_NET_WM_STATE_TOGGLE)
+                 (if (exwm-layout--fullscreen-p)
+                     (exwm-layout-unset-fullscreen id)
+                   (exwm-layout-set-fullscreen id)
+                   (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new)))))
+        ;; _NET_WM_STATE_DEMANDS_ATTENTION
+        ;; FIXME: check (may require other properties set)
+        (when (memq xcb:Atom:_NET_WM_STATE_DEMANDS_ATTENTION props)
+          (when (= action xcb:ewmh:_NET_WM_STATE_ADD)
+            (unless (eq exwm--frame exwm-workspace--current)
+              (set-frame-parameter exwm--frame 'exwm-urgency t)
+              (setq exwm-workspace--switch-history-outdated t)))
+          ;; xcb:ewmh:_NET_WM_STATE_REMOVE?
+          ;; xcb:ewmh:_NET_WM_STATE_TOGGLE?
+          )
+        (xcb:+request exwm--connection
+            (make-instance 'xcb:ewmh:set-_NET_WM_STATE
+                           :window id :data (vconcat props-new)))
+        (xcb:flush exwm--connection)))))
+
+(defun exwm--on-wm-protocols (_id data)
+  "Handle WM_PROTOCOLS message with DATA."
+  (let ((type (elt data 0)))
+    (cond ((= type xcb:Atom:_NET_WM_PING)
+           (setq exwm-manage--ping-lock nil))
+          (t (exwm--log "Unhandled WM_PROTOCOLS of type: %d" type)))))
+
+(defun exwm--on-wm-change-state (id data)
+  "Handle WM_CHANGE_STATE message with ID and DATA."
+  (let ((buffer (exwm--id->buffer id)))
+    (when (and (buffer-live-p buffer)
+               (= (elt data 0) xcb:icccm:WM_STATE:IconicState))
+      (with-current-buffer buffer
+        (if exwm--floating-frame
+            (call-interactively #'exwm-floating-hide)
+          (bury-buffer))))))
+
 (defun exwm--on-ClientMessage (raw-data _synthetic)
   "Handle ClientMessage event.
 RAW-DATA contains unmarshalled ClientMessage event data."
-  (let ((obj (make-instance 'xcb:ClientMessage))
-        type id data)
-    (xcb:unmarshal obj raw-data)
-    (setq type (slot-value obj 'type)
-          id (slot-value obj 'window)
-          data (slot-value (slot-value obj 'data) 'data32))
-    (exwm--log "atom=%s(%s) id=#x%x data=%s" (x-get-atom-name type exwm-workspace--current)
-               type (or id 0) data)
-    (cond
-     ;; _NET_NUMBER_OF_DESKTOPS.
-     ((= type xcb:Atom:_NET_NUMBER_OF_DESKTOPS)
-      (let ((current (exwm-workspace--count))
-            (requested (elt data 0)))
-        ;; Only allow increasing/decreasing the workspace number by 1.
-        (cond
-         ((< current requested)
-          (make-frame))
-         ((and (> current requested)
-               (> current 1))
-          (let ((frame (car (last exwm-workspace--list))))
-            (delete-frame frame))))))
-     ;; _NET_CURRENT_DESKTOP.
-     ((= type xcb:Atom:_NET_CURRENT_DESKTOP)
-      (exwm-workspace-switch (elt data 0)))
-     ;; _NET_ACTIVE_WINDOW.
-     ((= type xcb:Atom:_NET_ACTIVE_WINDOW)
-      (let ((buffer (exwm--id->buffer id))
-            window)
-        (if (buffer-live-p buffer)
-          ;; Either an `exwm-mode' buffer (an X window) or a floating frame.
-          (with-current-buffer buffer
-            (when (eq exwm--frame exwm-workspace--current)
-              (if exwm--floating-frame
-                  (select-frame exwm--floating-frame)
-                (setq window (get-buffer-window nil t))
-                (unless window
-                  ;; State change: iconic => normal.
-                  (setq window (frame-selected-window exwm--frame))
-                  (set-window-buffer window (current-buffer)))
-                ;; Focus transfer.
-                (select-window window))))
-          ;; A workspace.
-          (dolist (f exwm-workspace--list)
-            (when (eq id (frame-parameter f 'exwm-outer-id))
-              (x-focus-frame f t))))))
-     ;; _NET_CLOSE_WINDOW.
-     ((= type xcb:Atom:_NET_CLOSE_WINDOW)
-      (let ((buffer (exwm--id->buffer id)))
-        (when (buffer-live-p buffer)
-          (exwm--defer 0 #'kill-buffer buffer))))
-     ;; _NET_WM_MOVERESIZE
-     ((= type xcb:Atom:_NET_WM_MOVERESIZE)
-      (let ((direction (elt data 2))
-            (buffer (exwm--id->buffer id)))
-        (unless (and buffer
-                     (not (buffer-local-value 'exwm--floating-frame buffer)))
-          (cond ((= direction
-                    xcb:ewmh:_NET_WM_MOVERESIZE_SIZE_KEYBOARD)
-                 ;; FIXME
-                 )
-                ((= direction
-                    xcb:ewmh:_NET_WM_MOVERESIZE_MOVE_KEYBOARD)
-                 ;; FIXME
-                 )
-                ((= direction xcb:ewmh:_NET_WM_MOVERESIZE_CANCEL)
-                 (exwm-floating--stop-moveresize))
-                ;; In case it's a workspace frame.
-                ((and (not buffer)
-                      (catch 'break
-                        (dolist (f exwm-workspace--list)
-                          (when (or (eq id (frame-parameter f 'exwm-outer-id))
-                                    (eq id (frame-parameter f 'exwm-id)))
-                            (throw 'break t)))
-                        nil)))
-                (t
-                 ;; In case it's a floating frame,
-                 ;; move the corresponding X window instead.
-                 (unless buffer
-                   (catch 'break
-                     (dolist (pair exwm--id-buffer-alist)
-                       (with-current-buffer (cdr pair)
-                         (when
-                             (and exwm--floating-frame
-                                  (or (eq id
-                                          (frame-parameter exwm--floating-frame
-                                                           'exwm-outer-id))
-                                      (eq id
-                                          (frame-parameter exwm--floating-frame
-                                                           'exwm-id))))
-                           (setq id exwm--id)
-                           (throw 'break nil))))))
-                 ;; Start to move it.
-                 (exwm-floating--start-moveresize id direction))))))
-     ;; _NET_REQUEST_FRAME_EXTENTS
-     ((= type xcb:Atom:_NET_REQUEST_FRAME_EXTENTS)
-      (let ((buffer (exwm--id->buffer id))
-            top btm)
-        (if (or (not buffer)
-                (not (buffer-local-value 'exwm--floating-frame buffer)))
-            (setq top 0
-                  btm 0)
-          (setq top (window-header-line-height)
-                btm (window-mode-line-height)))
-        (xcb:+request exwm--connection
-            (make-instance 'xcb:ewmh:set-_NET_FRAME_EXTENTS
-                           :window id
-                           :left 0
-                           :right 0
-                           :top top
-                           :bottom btm)))
-      (xcb:flush exwm--connection))
-     ;; _NET_WM_DESKTOP.
-     ((= type xcb:Atom:_NET_WM_DESKTOP)
-      (let ((buffer (exwm--id->buffer id)))
-        (when (buffer-live-p buffer)
-          (exwm-workspace-move-window (elt data 0) id))))
-     ;; _NET_WM_STATE
-     ((= type xcb:Atom:_NET_WM_STATE)
-      (let ((action (elt data 0))
-            (props (list (elt data 1) (elt data 2)))
-            (buffer (exwm--id->buffer id))
-            props-new)
-        ;; only support _NET_WM_STATE_FULLSCREEN / _NET_WM_STATE_ADD for frames
-        (when (and (not buffer)
-                   (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
-                   (= action xcb:ewmh:_NET_WM_STATE_ADD))
-          (xcb:+request
-              exwm--connection
-              (make-instance 'xcb:ewmh:set-_NET_WM_STATE
-                             :window id
-                             :data (vector xcb:Atom:_NET_WM_STATE_FULLSCREEN)))
-          (xcb:flush exwm--connection))
-        (when buffer                    ;ensure it's managed
-          (with-current-buffer buffer
-            ;; _NET_WM_STATE_FULLSCREEN
-            (when (or (memq xcb:Atom:_NET_WM_STATE_FULLSCREEN props)
-                      (memq xcb:Atom:_NET_WM_STATE_ABOVE props))
-              (cond ((= action xcb:ewmh:_NET_WM_STATE_ADD)
-                     (unless (exwm-layout--fullscreen-p)
-                       (exwm-layout-set-fullscreen id))
-                     (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new))
-                    ((= action xcb:ewmh:_NET_WM_STATE_REMOVE)
-                     (when (exwm-layout--fullscreen-p)
-                       (exwm-layout-unset-fullscreen id)))
-                    ((= action xcb:ewmh:_NET_WM_STATE_TOGGLE)
-                     (if (exwm-layout--fullscreen-p)
-                         (exwm-layout-unset-fullscreen id)
-                       (exwm-layout-set-fullscreen id)
-                       (push xcb:Atom:_NET_WM_STATE_FULLSCREEN props-new)))))
-            ;; _NET_WM_STATE_DEMANDS_ATTENTION
-            ;; FIXME: check (may require other properties set)
-            (when (memq xcb:Atom:_NET_WM_STATE_DEMANDS_ATTENTION props)
-              (when (= action xcb:ewmh:_NET_WM_STATE_ADD)
-                (unless (eq exwm--frame exwm-workspace--current)
-                  (set-frame-parameter exwm--frame 'exwm-urgency t)
-                  (setq exwm-workspace--switch-history-outdated t)))
-              ;; xcb:ewmh:_NET_WM_STATE_REMOVE?
-              ;; xcb:ewmh:_NET_WM_STATE_TOGGLE?
-              )
-            (xcb:+request exwm--connection
-                (make-instance 'xcb:ewmh:set-_NET_WM_STATE
-                               :window id :data (vconcat props-new)))
-            (xcb:flush exwm--connection)))))
-     ((= type xcb:Atom:WM_PROTOCOLS)
-      (let ((type (elt data 0)))
-        (cond ((= type xcb:Atom:_NET_WM_PING)
-               (setq exwm-manage--ping-lock nil))
-              (t (exwm--log "Unhandled WM_PROTOCOLS of type: %d" type)))))
-     ((= type xcb:Atom:WM_CHANGE_STATE)
-      (let ((buffer (exwm--id->buffer id)))
-        (when (and (buffer-live-p buffer)
-                   (= (elt data 0) xcb:icccm:WM_STATE:IconicState))
-          (with-current-buffer buffer
-            (if exwm--floating-frame
-                (call-interactively #'exwm-floating-hide)
-              (bury-buffer))))))
-     (t
-      (exwm--log "Unhandled: %s(%d)"
-                 (x-get-atom-name type exwm-workspace--current) type)))))
+  (let* ((obj (let ((m (make-instance 'xcb:ClientMessage)))
+                (xcb:unmarshal m raw-data)
+                m))
+         (type (slot-value obj 'type))
+         (id (slot-value obj 'window))
+         (data (slot-value (slot-value obj 'data) 'data32))
+         (fn (alist-get type exwm--client-message-functions)))
+    (if (not fn)
+        (exwm--log "Unhandled: %s(%d)"
+                   (x-get-atom-name type exwm-workspace--current) type)
+      (exwm--log "atom=%s(%s) id=#x%x data=%s"
+                 (x-get-atom-name type exwm-workspace--current)
+                 type (or id 0) data)
+      (funcall fn id data))))
 
 (defun exwm--on-SelectionClear (data _synthetic)
   "Handle SelectionClear events.
@@ -888,7 +908,6 @@ manager.  If t, replace it, if nil, abort and ask the user if `ask'."
         (xcb:+request exwm--connection se))
       (setq exwm--wmsn-window new-owner))))
 
-;;;###autoload
 (cl-defun exwm-init (&optional frame)
   "Initialize EXWM.
 FRAME, if given, indicates the X display EXWM should manage."
@@ -918,6 +937,18 @@ FRAME, if given, indicates the X display EXWM should manage."
         ;; Initialize ICCCM/EWMH support
         (xcb:icccm:init exwm--connection t)
         (xcb:ewmh:init exwm--connection t)
+        (setq
+         exwm--client-message-functions
+         (list (cons xcb:Atom:_NET_NUMBER_OF_DESKTOPS #'exwm--on-net-number-of-desktops)
+               (cons xcb:Atom:_NET_CURRENT_DESKTOP #'exwm--on-net-current-desktop)
+               (cons xcb:Atom:_NET_ACTIVE_WINDOW #'exwm--on-net-active-window)
+               (cons xcb:Atom:_NET_CLOSE_WINDOW #'exwm--on-net-close-window)
+               (cons xcb:Atom:_NET_REQUEST_FRAME_EXTENTS
+                     #'exwm--on-net-request-frame-extents)
+               (cons xcb:Atom:_NET_WM_DESKTOP #'exwm--on-net-wm-desktop)
+               (cons xcb:Atom:_NET_WM_STATE #'exwm--on-net-wm-state)
+               (cons xcb:Atom:WM_PROTOCOLS #'exwm--on-wm-protocols)
+               (cons xcb:Atom:WM_CHANGE_STATE #'exwm--on-wm-change-state)))
         ;; Try to register window manager selection.
         (exwm--wmsn-acquire exwm-replace)
         (when (xcb:+request-checked+request-check exwm--connection
@@ -954,7 +985,6 @@ FRAME, if given, indicates the X display EXWM should manage."
      (warn "[EXWM] EXWM fails to start (%s: %s)" (car err) (cdr err)))))
 
 
-;;;###autoload
 (defun exwm-exit ()
   "Exit EXWM."
   (interactive)
@@ -972,11 +1002,15 @@ FRAME, if given, indicates the X display EXWM should manage."
     (xcb:disconnect exwm--connection))
   (setq exwm--connection nil)
   (setq exwm--terminal nil)
+  (setenv "INSIDE_EXWM" nil)
   (exwm--log "Exited"))
 
 ;;;###autoload
 (defun exwm-enable (&optional undo)
-  "Enable/Disable EXWM."
+  "Enable/Disable EXWM.
+Optional argument UNDO may be either of the following symbols:
+- `undo' prevents reinitialization.
+- `undo-all' attempts to revert all hooks and advice."
   (exwm--log "%s" undo)
   (pcase undo
     (`undo                              ;prevent reinitialization
@@ -990,7 +1024,9 @@ FRAME, if given, indicates the X display EXWM should manage."
        (advice-remove i #'exwm--server-eval-at)))
     (_                                  ;enable EXWM
      (setq frame-resize-pixelwise t     ;mandatory; before init
-           window-resize-pixelwise t)
+           window-resize-pixelwise t
+           x-no-window-manager t)
+     (setenv "INSIDE_EXWM" "1")
      ;; Ignore unrecognized command line arguments.  This can be helpful
      ;; when EXWM is launched by some session manager.
      (push #'vector command-line-functions)
@@ -1062,7 +1098,7 @@ FUNCTION is the function to be evaluated, ARGS are the arguments."
   ;; This is invoked instead of `save-buffers-kill-emacs' (C-x C-c) on client
   ;; frames.
   (if (exwm--terminal-p)
-      (exwm--confirm-kill-emacs "Kill terminal?")
+      (exwm--confirm-kill-emacs "Kill terminal")
     t))
 
 (defun exwm--confirm-kill-emacs (prompt &optional force)
@@ -1087,7 +1123,7 @@ If FORCE is any other non-nil value, force killing of Emacs."
             (`break (y-or-n-p prompt))
             (x x)))
          (t
-          (yes-or-no-p (format "[EXWM] %d X window(s) will be destroyed.  %s"
+          (yes-or-no-p (format "[EXWM] %d X window(s) will be destroyed.  %s?"
                                (length exwm--id-buffer-alist) prompt))))
     ;; Run `kill-emacs-hook' (`server-force-stop' excluded) before Emacs
     ;; frames are unmapped so that errors (if any) can be visible.
@@ -1103,8 +1139,5 @@ If FORCE is any other non-nil value, force killing of Emacs."
     ;; Set the return value.
     t))
 
-
-
 (provide 'exwm)
-
 ;;; exwm.el ends here
